/*
	midi player
	(c) 1999-2002 BERO <bero@geocities.co.jp>
	under GPL or notify me
*/

#define _arch_dreamcast
#include <kos.h>
#include "aica.h"

#include "patch.h"

#define	ENDIAN(a)	(((a)<<24)|(((a)&0xff00)<<8)|(((a)>>8)&0xff00)|((a)>>24))
#define	GETLONG(p)	(((p)[0]<<24)|((p)[1]<<16)|((p)[2]<<8)|(p)[3])
#define	GETSHORT(p)	(((p)[0]<<8)|(p)[1])

#define	Printf(args...)	
//#define	Printf	printf

static int _memcmp(const char *dst,const char *src,int len)
{
	while(len--) {
		int cmp = *dst++ - *src++;
		if (cmp) return cmp;
	}
	return 0;
}
#define	memcmp	_memcmp

typedef struct {
	unsigned char *start;
	int size;
	unsigned char *pointer;
	int	count;
	int stop;
	unsigned char lastmsg;
} TRACK;
#define	MAXTRACK	32

typedef struct {
	void *data;
	int ntrack; /* number of tracks */
	int delta;	/* 4delta time */
	int ndelta;
	int count;	/* JE^ */
	int count_low;
	int loop;
	int playing;
	TRACK track[MAXTRACK];
} MIDIBUF;

typedef struct {
	int note,prog;
	int pitch;
	int volume;
	int pan;
	int mvol,expression;
} MIDI_CH;

static MIDI_CH midi_ch[16];

#define	MAXVOICE	64
static int start_voice = 0;
static int maxvoice = 63;
static int voice[MAXVOICE];
static SAMPLE *voice_sample[MAXVOICE];

#define	GET_DELTA(delta,p)	{int c; delta=0; do { c=*p++; delta=(delta<<7)|(c&0x7f);} while(c>0x7f); }

#define	TEMPO2NDELTA(tempo,base)	((long long)(1000000*256/60)*(base)/(tempo))


static void ctrl_change(int ch,int v1,int v2);
static void prog_change(int ch,int prog);
static void note_on(int ch,int note,int vol);
static void note_off(int ch,int note);
static void pitch_bend(int ch,int pitch);

static int parse_track(TRACK *track,MIDIBUF *midi)
{
	while(!track->stop && track->count<=midi->count) {
		unsigned char *p = track->pointer;
		int msg,ch;
		Printf("now=%d pointer=%d\n",midi->count,track->count);

		msg = *p++;
		if (msg<0x80) {
			msg = track->lastmsg; p--;
		}
		track->lastmsg = msg;
		Printf("%02x:",msg);
		ch = msg&0xf;
		switch(msg&0xf0) {
		case 0x90:
			Printf("note on:%d %d\n",p[0],p[1]);
			note_on(ch,p[0],p[1]);
			p+=2;
			break;
		case 0x80:
			Printf("note off:%d %d\n",p[0],p[1]);
			note_off(ch,p[0]);
			p+=2;
			break;
		case 0xa0:
			Printf("porihonic pressure:%d %d\n",p[0],p[1]);
			p+=2;
		case 0xb0:
			Printf("ctrl change:%d %d\n",p[0],p[1]);
			ctrl_change(ch,p[0],p[1]);
			p+=2;
			break;
		case 0xc0:
			Printf("prog change:%d\n",p[0]);
			prog_change(ch,p[0]);
			p+=1;
			break;
		case 0xd0:
			Printf("channel pressure:%d\n",p[1]);
			p+=1;
			break;
		case 0xe0:
			Printf("pitch bend:%d\n",p[0]|(p[1]<<7));
			pitch_bend(ch,p[0]|(p[1]<<7));
			p+=2;
			break;
		case 0xf0:
			switch(msg){
			case 0xff:
				{ int len,type;
				type = *p++;
				Printf("meta:%x ",type);
				GET_DELTA(len,p);

				switch(type) {
				case 0x51:
					/* tempo */
					{int tempo = (p[0]<<16)|(p[1]<<8)|p[2];
					Printf("tempo=%d",60000000/tempo);
					midi->ndelta = TEMPO2NDELTA(tempo,midi->delta);
					Printf("ndelta:%d\n",midi->ndelta);
					p+=3;
					}
					break;
				case 0x2f:
					Printf("end of track\n");
					/* end of track */
					track->stop = 1;
				default:
					//while (len--) Printf(" %x",*p++);
					p+=len;
					break;
				}
				}
				break;
				
			case 0xf0:
			case 0xf7:
				Printf("exclusive:");
				{int len;
				GET_DELTA(len,p);
				p+=len;
				}
				break;
			case 0xf1:
				Printf("quota frame:%x\n",p[0]);
				p+=1;
				break;
			case 0xf2:
				Printf("song position:%d\n",p[0]|(p[1]<<7));
				p+=2;
				break;
			case 0xf3:
				Printf("song select:%x\n",p[0]);
				p+=1;
				break;
			default:
				Printf("other\n");
			}
			break;
		default:
			Printf("unknown\n");
			break;
		}

		/* next delta */
		{
			int delta;
			GET_DELTA(delta,p);
			track->count+=delta;
			Printf("delta:%d\n",delta);
		}

		track->pointer = p;
	}
	return 0;
}

int play_init(MIDIBUF *midi)
{
	int i;
	TRACK *track;

	Printf("init\n");
	memset(midi_ch,0,sizeof(midi_ch));
	for(i=0;i<16;i++) {
		midi_ch[i].mvol = 127;
		midi_ch[i].expression = 127;
		midi_ch[i].volume = 127;
	}

	midi->count = midi->count_low=0;
	midi->ndelta = TEMPO2NDELTA(500000,midi->delta);
	midi->playing = 1;

	Printf("%d %d %d\n",midi->count,midi->delta,midi->ndelta);

	for(i=0,track=midi->track;i<midi->ntrack;i++,track++) {
		track->pointer = track->start;
		track->count = 0;
		track->stop = 0;

		/* first delta */
		{
			unsigned char *p = track->pointer;
			int delta;
			GET_DELTA(delta,p);
			track->count+=delta;
			track->pointer = p;
		}
	}

	return 0;
}

void playback(MIDIBUF *midi)
{
	int i;
	int low;
	int stop_tracks;

	if (midi->data==NULL) return;
	low = midi->count_low+midi->ndelta;
	midi->count+=low/256;
	midi->count_low=low&255;
	stop_tracks = 0;

	for(i=0;i<midi->ntrack;i++) {
		parse_track(&midi->track[i],midi);
		stop_tracks +=midi->track[i].stop;
	}
//	printf("%d ",midi->count);

	if (midi->ntrack==stop_tracks) {
		if (midi->loop)
			play_init(midi);
		else
			midi->playing = 0;
	}
}

int midi_set(unsigned char *buf,MIDIBUF *midi)
{
	unsigned char *bufp;
	int format,ntrack,deltacnt,chunksize;
	int track;

	midi->data = buf;
	bufp = buf;

	if (memcmp(bufp,"MThd",4)) {
		Printf("not midi");
		return -1;
	}
	chunksize = GETLONG(bufp+4);
	format = GETSHORT(bufp+8);
	ntrack = GETSHORT(bufp+10);
	deltacnt = GETSHORT(bufp+12);
	bufp+=chunksize+8;
	Printf("format:%d track:%d deltacnt:%d\n",format,ntrack,deltacnt);
	midi->ntrack = ntrack;
	midi->delta = deltacnt;

	for(track=0;track<ntrack;track++){

		Printf("track:%d\n",track);
		if (memcmp(bufp,"MTrk",4)) {
			char tmp[5];
			memcpy(tmp,bufp,4);
			tmp[4] = 0;
			Printf("unknouwn chunk:%s\n",tmp);
			return -1;
		}
		chunksize = GETLONG(bufp+4);
		Printf("chunksize:%d\n",chunksize);
		bufp+=8;
		midi->track[track].start = bufp;
		midi->track[track].size = chunksize;
		bufp+=chunksize;
	}

	return 0;
}


static int alloc_voice(int ch,int note)
{
	int n = ch*256+note;
	int i;
//	int mask = SPU_ActiveVoice();
//	printf("%x ",~mask);
	for(i=start_voice;i<maxvoice;i++) {
		if (voice[i]==0 /*&& (mask&1)*/) {
			voice[i]=n;
			return i;
		}
//		mask>>=1;
	}
	printf("can't alloc");
	return -1;
}

static int free_voice(int ch,int note)
{
	int n = ch*256+note;
	int i;
	for(i=start_voice;i<maxvoice;i++) {
		if (voice[i]==n) {
			voice[i]=0;
			return i;
		}
	}
	return -1;
}

void prog_change(int ch,int prog)
{
	midi_ch[ch].prog=prog;
}

static int calc_freq(int note,int pitch);

void pitch_bend(int ch,int pitch)
{
	int i;

	midi_ch[ch].pitch = pitch-8192;
	if (ch==9) return;

	for(i=start_voice;i<maxvoice;i++) {
		if (voice[i] && ((voice[i]>>8)==ch)) {
			SAMPLE *sample = voice_sample[i];
			int freq;
			freq = calc_freq(midi_ch[ch].note,midi_ch[ch].pitch);
			freq = (long long)freq*sample->freq/(sample->base_freq);
			aica_freq(i,freq);
		}
	}
}

void ctrl_change(int ch,int v1,int v2)
{
	switch(v1) {
	case 0x00: /* bank select */
	case 0x07: midi_ch[ch].mvol = v1; break;
	case 0x0b: midi_ch[ch].expression = v1; break;
	case 0x0a: midi_ch[ch].pan = v2; break;
	}
}

void note_off(int ch,int note)
{
	int n;

	n = free_voice(ch,note);
	if (n<0) return;
//	n = ch;

	aica_stop(n);
}

static SAMPLE *search_sample(PATCH *patch,int freq);
static int master_vol = 127;

void note_on(int ch,int note,int vol)
{
	int n,freq;
	SAMPLE *sample;
	PATCH *patch;

	if (vol==0) {
		note_off(ch,note);
		return;
	}

	midi_ch[ch].note = note;

	n = alloc_voice(ch,note);
	if (n<0) return;
//	n = ch;

	if (ch==9) {
		patch = patches[1][note];
		sample = patch->sample[0];
		freq = sample->freq;
	} else {
		patch = patches[0][midi_ch[ch].prog];
	//	patch = patches[0][0];
		freq = calc_freq(note,midi_ch[ch].pitch);
		sample = search_sample(patch,freq);
		freq = (long long)freq*sample->freq/(sample->base_freq);
	}

	voice_sample[n] = sample;
	vol = (vol*master_vol*midi_ch[ch].mvol*midi_ch[ch].expression)>>(7*3);
	aica_play(n,sample->mode,sample->data,sample->loop_start,sample->loop_end,freq,vol,128/* pan */,sample->loop_flag);
}

const static unsigned int ftbl[] = {
	/* freq*100 */
	/* C C# D D# E F F# G G# A A# B */
    8175,    8661,    9177,    9722,   10300,   10913,   11562,   12249,   12978,   13750,   14567,   15433,
   16351,   17323,   18354,   19445,   20601,   21826,   23124,   24499,   25956,   27500,   29135,   30867,
   32703,   34647,   36708,   38890,   41203,   43653,   46249,   48999,   51913,   55000,   58270,   61735,
   65406,   69295,   73416,   77781,   82406,   87307,   92498,   97998,  103826,  110000,  116540,  123470,
  130812,  138591,  146832,  155563,  164813,  174614,  184997,  195997,  207652,  220000,  233081,  246941,
  261625,  277182,  293664,  311127,  329627,  349228,  369994,  391995,  415304,  440000,  466163,  493883,
  523251,  554365,  587329,  622254,  659255,  698456,  739988,  783991,  830609,  880000,  932327,  987766,
 1046502, 1108730, 1174659, 1244508, 1318510, 1396913, 1479977, 1567982, 1661219, 1760000, 1864655, 1975533,
 2093005, 2217461, 2349318, 2489016, 2637020, 2793826, 2959955, 3135964, 3322438, 3520000, 3729311, 3951067,
 4186010, 4434923, 4698637, 4978032, 5274041, 5587652, 5919911, 6271928, 6644876, 7040001, 7458621, 7902133,
 8372019, 8869845, 9397273, 9956064,10548082,11175304,11839822,12543854,13289750,14080000,
};

static int calc_freq(int note,int pitch)
{
	int freq;

	while (pitch<0) {
		note--;
		pitch+=4096;
	}
	note += pitch>>12;
	pitch&= 4096-1;

	// freq = ((f1*(4096-pitch)) + (f2*pitch)) / 4096;
	freq = ftbl[note] + (ftbl[note+1]-ftbl[note])*pitch/4096;
	return freq;
}

static inline int ABS(int a) {return a>=0?a:-a; }
static inline int MIN(int a,int b) {return a<b?a:b; }

static SAMPLE *search_sample(PATCH *patch,int freq)
{
	SAMPLE *sample;

	if (patch->n_samples == 1) {
		return patch->sample[0];
	} else {
		int i,best_diff = 0x7fffffff;
		SAMPLE *s;
		for(i=0;i<patch->n_samples;i++) {
			s = patch->sample[i];
			if (s->low_freq <= freq && freq<= s->high_freq) {
				return s;
				break;
			} else {
				int diff = MIN(ABS(freq - s->low_freq),ABS(freq-s->high_freq));
				if (diff<best_diff) {
					best_diff = diff;
					sample = s;
				}
			}
		}
		return sample;
	}
}

/* API */
static MIDIBUF midi[1];

void midi_alloc_voice(unsigned start,unsigned num)
{
	start_voice = start;
	maxvoice = start+num;
}

int midi_play(void *buf,int loop)
{
	midi_stop();
	if (midi_set(buf,midi)) return -1;
	midi[0].loop = loop;
	play_init(midi);
}

void midi_callback(void)
{
	if (midi[0].playing) {
		playback(midi);
	}
}

void midi_stop(void)
{
	int i;
	midi[0].playing = 0;
	for(i=start_voice;i<maxvoice;i++) {
		aica_stop(i);
	}
}

int midi_isplay(void)
{
	return midi[0].playing;
}

void midi_free(void)
{
	midi_stop();
	if (midi[0].data) {
		free(midi[0].data);
		midi[0].data = 0;
	}
}

void midi_volume(int vol)
{
	master_vol = vol;
}

static void vsync_handler(unsigned int code)
{
	midi_callback();
	/* default handler */
	maple_vbl_irq_hnd(code);
}

void midi_install_handler(void)
{
	asic_evt_set_handler(ASIC_EVT_PVR_VBLINT, vsync_handler);
}
