#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>

#define	DC

#ifdef DC
#define _arch_dreamcast
#include <kos.h>

#define	_open	fs_open
#define	_close	fs_close
#define	_read	fs_read
#define	_lseek	fs_seek
#define	ERR_HANDLE	0

#else

#define	ERR_HANDLE	-1

#endif

#include "patch.h"
#include "midi.h"

#define	SM_8BIT		1
#define	SM_16BIT	0
#define	SM_ADPCM	2

#define	Printf(args...)

void pcm16_to_adpcm(unsigned char *dst,const short *src,size_t length,int sign);
void pcm8_to_adpcm(unsigned char *dst,const signed char *src,size_t length,int sign);

static unsigned int spu_base = 16;
static unsigned int spu_end = 2*1024*1024;

unsigned int spu_malloc(size_t size)
{
	int ret = spu_base;
	size = (size+3)&~3;
	if (spu_base+size>=spu_end) {
		dbgio_printf("spumem overflow");
		exit(-1);
	}
	spu_base += size;
	return ret;
}

void midi_alloc_mem(unsigned start,unsigned size)
{
	spu_base = start;
	spu_end = start+size;
}

/* GUS */

typedef unsigned char byte;
typedef byte byte2[2];
typedef byte byte4[4];

typedef struct {
	char id[22];
	char description[60];
	byte n_inst;
	byte voices;
	byte channels;
	byte2 waveforms; /* le */
	byte2 master_vol;
	byte4 data_size;
	byte reserved[36];
	byte2 inst_id;
	char inst_name[16];
	byte4 inst_size;
	byte layers;
	byte reserved2[40];
	byte layer_duplicate;
	byte layer;
	byte4 layer_size;
	byte n_samples;
	byte reserved3[40];
} patchhdr;

typedef struct {
	char name[7];
	byte fractions;
	byte4 size;
	byte4 loop_start;
	byte4 loop_end;
	byte2 rate;
	byte4 low_freq;
	byte4 high_freq;
	byte4 root_freq;
	byte2 tune;
	byte pan;
	byte envelope[4][3];
	byte tremolo_sweep,tremolo_rate,tremolo_depth;
	byte vibrato_sweep,vibrate_rate,vibrate_depth;
	byte mode;
	byte2 scale_freq;
	byte2 scale_factor;
	char reserved[36];
} samplehdr;


static int read_long(byte4 p)
{
	return p[0] | (p[1]<<8) | (p[2]<<16) | (p[3]<<24);
}

static int read_short(byte2 p)
{
	return p[0] | (p[1]<<8);
}

static int total,total2;

PATCH* load_patch(const char *file,int load_mode)
{
//	FILE *fp = fopen(file,"rb");
	int fd = _open(file,O_RDONLY);

	patchhdr hdr;
	PATCH *patch;
	int i;

	if (fd==ERR_HANDLE) return NULL;
	printf("file:%s\n",file);

	_read(fd,&hdr,sizeof(hdr));

	if (memcmp(hdr.id,"GF1PATCH110\0ID#000002",22)!=0
	&&  memcmp(hdr.id,"GF1PATCH100\0ID#000002",22)!=0) return NULL;

	hdr.inst_name[15] = 0;
	Printf("%s inst=%d layers=%d vol=%d\n",hdr.inst_name,hdr.n_inst,hdr.layers,read_short(hdr.master_vol));

	patch = malloc(sizeof(PATCH)+sizeof(SAMPLE*)*(hdr.n_samples-1));
	patch->n_samples = hdr.n_samples;

	for(i=0;i<hdr.n_samples;i++) {
		int mode,j,nsamples,size;
		void *buf,*buf2;
		SAMPLE *sample;

		samplehdr shdr;
		char name[8];

		sample = malloc(sizeof(SAMPLE));
		patch->sample[i] = sample;

		_read(fd,&shdr,sizeof(shdr));
		sample->size = read_long(shdr.size);
		sample->loop_start = read_long(shdr.loop_start);
		sample->loop_end = read_long(shdr.loop_end)+1;
		sample->freq = read_short(shdr.rate);
		sample->low_freq = read_long(shdr.low_freq);
		sample->high_freq = read_long(shdr.high_freq);
		sample->base_freq = read_long(shdr.root_freq);
		mode = shdr.mode;


		name[7] = 0;
		memcpy(name,shdr.name,7);

		Printf("%s size=%d start=%d end=%d rate=%d low=%d high=%d root=%d"
		" mode=%x scale_freq=%d scale_factor=%d\n",name
		,sample->size,sample->loop_start,sample->loop_end,sample->freq
		,sample->low_freq,sample->high_freq,sample->base_freq, shdr.mode
		,read_short(shdr.scale_freq),read_short(shdr.scale_factor)
		);
		Printf("%ssigned %sbit %s %s %s\n", (mode&2)?"un":"",(mode&1)?"16":"8",
		(mode&4)?"loop":"noloop",(mode&8)?"pingpomg":"",(mode&16)?"reverse":"");

		sample->loop_flag = (mode&4)!=0;

		if (mode&1) {
			sample->mode = SM_16BIT;
			sample->loop_start>>=1;
			sample->loop_end  >>=1;
		} else {
			sample->mode = SM_8BIT;
		}

		size = sample->size;


		nsamples = size;
		if (mode&1) nsamples>>=1;

		total2+=nsamples/2;
		total+=size;

		if (nsamples>65534) Printf("#over\n");

		buf = malloc(size);
		_read(fd,buf,size);

	  if (load_mode==LOAD_ADPCM) {
		/* USE ADPCM */
		sample->size = (nsamples+1)/2;
		sample->mode = SM_ADPCM;
		buf2 = malloc((nsamples+1)/2);
		switch(mode&3) {
		case 0: /* 8bit signed */
			pcm8_to_adpcm(buf2,buf,size,0); break;
		case 2: /* 8bit unsigned */
			pcm8_to_adpcm(buf2,buf,size,~0x7f); break;
		case 1: /* 16bit signed */
			pcm16_to_adpcm(buf2,buf,size,0); break;
		case 3: /* 16bit unsigned */
			pcm16_to_adpcm(buf2,buf,size,~0x7fff); break;
		}
		free(buf);
		sample->data = buf2;
	  } else if (load_mode==LOAD_8BIT) {
		/* 8BIT PCM */
		sample->size = nsamples;
		sample->mode = SM_8BIT;
		switch(mode&3) {
		case 0: /* 8bit signed */
			break;
		case 2: /* 8bit unsigned */
			for(j=0;j<nsamples;j++) ((char*)buf)[j]^=~0x7f;
			break;
		case 1: /* 16bit signed */
			for(j=0;j<nsamples;j++) ((char*)buf)[j] = (((short*)buf)[j])>>8;
			break;
		case 3: /* 16bit unsigned */
			for(j=0;j<nsamples;j++) ((char*)buf)[j] = (((short*)buf)[j]^~0x7fff)>>8;
			break;
		}
		sample->data = buf;
	  } else {
		switch(mode&3) {
		case 0: /* 8bit signed */
			for(j=0;j<size;j++) ((char*)buf)[j]^=~0x7f;
			break;
		case 3: /* 16bit unsigned */
			for(j=0;j<size;j+=2) *((short*)(buf+j))^=~0x7fff;
			break;
		}
		sample->data = buf;
	  }

#ifdef DC
		{
			int spu_buf = spu_malloc(sample->size);
			spu_memload(spu_buf,sample->data,sample->size);
			free(sample->data);
			sample->data = spu_buf;
		}
#endif

	//	free(buf);
	}
	_close(fd);

	return patch;
}

enum {PROG,DRUM,IGNORE};

#define	MAXFILE	256

PPTR patches[2][128];

int midi_load_patchs(const char *file,const char *basedir,int load_mode)
{
#define	MAXWORDS	10

	FILE *fp;
	char tmp[1024], *w[MAXWORDS];
	int mode = PROG;
	int bank = 0;

	struct {
		char name[16];
		void *patch;
	} loaded[MAXFILE];

	int n_loaded;

	if ((fp=fopen(file,"rt"))==NULL) return -1;

	memset(loaded,0,sizeof(loaded));
	n_loaded = 0;

	while (fgets(tmp, sizeof(tmp), fp)) {
		int words=0;
		char *w0;

		w[0] = w0 = strtok(tmp," \t\r\n\240");
		if (w0==NULL || (*w0=='#')) continue;
		while (w[words] && (words<MAXWORDS)) {
			w[++words]=strtok(NULL," \t\r\n\240");
		}
		if (w0[0]>='0' && w0[0]<='9') {
			int num = atoi(w0);
			int i;
			char *name;
			void *patch;

			/* currency bank 0 only */
			if (bank!=0) continue;

			name = w[1];
			for(i=0;i<n_loaded;i++) {
				if (strcmp(name,loaded[i].name)==0) break;
			}

			if (i!=n_loaded) {
				patch = loaded[i].patch;
			} else {
				char path[256];
				sprintf(path,"%s/%s.pat",basedir,name);
				patch = load_patch(path,load_mode);

				strcpy(loaded[n_loaded].name,name);
				loaded[n_loaded].patch = patch;
				n_loaded++;
			}
			patches[mode][num] = patch;
		} else if (!strcmp(w0,"bank")) {
			bank = atoi(w[1]);
			mode = PROG;
		} else if (!strcmp(w0,"drumset")) {
			bank = atoi(w[1]);
			mode = DRUM;
		}
	}
	fclose(fp);
	return 0;
}


